//reference System.dll
using System;
using System.Text;
using System.Collections.Generic;
using MCGalaxy.Network;
using MCGalaxy.Commands;

using MCGalaxy.DB;
using MCGalaxy.Drawing;
using MCGalaxy.Drawing.Brushes;
using MCGalaxy.Drawing.Ops;
using MCGalaxy.Maths;
using BlockID = System.UInt16;
using MCGalaxy.Undo;
using MCGalaxy.Util;

using Vector3 = MCGalaxy.Maths.Vec3F32;
using Vector3Int = MCGalaxy.Maths.Vec3S32;

namespace MCGalaxy
{
    public class CmdIO : Command2
    {
        public override string name { get { return "IO"; } }
        public override string shortcut { get { return "ImportObj"; } }
        public override bool MessageBlockRestricted { get { return true; } }
        public override string type { get { return "other"; } }
        public override bool museumUsable { get { return false; } }
        public override LevelPermission defaultRank { get { return LevelPermission.Admin; } }
        const string CONFIRMED_STRING = " confirm";
        public static bool printShit;
        public static Player omfg = null;
        public override void Use(Player p, string message, CommandData data) {  
            printShit = true;
            omfg = p;
            bool canUse = false;
            if (LevelInfo.IsRealmOwner(p.name, p.level.name)) canUse = true;
            if (p.group.Permission >= LevelPermission.Operator && p.group.Permission >= p.level.BuildAccess.Min) { canUse = true; }
            if (!canUse) {
                p.Message("&cYou can only use this command on your own maps."); return;
            }
            
            if (message == "") { Help(p); return; }
            bool confirmed = false;
            if (message.ToLower().EndsWith(CONFIRMED_STRING)) { confirmed = true; message = message.Substring(0, message.Length-CONFIRMED_STRING.Length); }

            string[] args      = message.SplitSpaces();
            string objUrl      = args[0];
            string scaleString = args.Length < 2 ? "16" : args[1];
            string textureUrl  = args.Length < 3 ? ""   : args[2];
            string paletteName = args.Length < 4 ? (p.level.name == "grief" ? "inf" : "color") : args[3];
            
            float scale = 16;
            if (!CommandParser.GetReal(p, scaleString, "scale", ref scale)) return;

            byte[] raw = HttpUtil.DownloadData(objUrl, p);
            if (raw == null) return;
            Obj obj = new Obj(Encoding.ASCII.GetString(raw), scale);
            
            //if (!confirmed) {
            //    Vector3Int minMark = new Vector3Int(p.level.Width/2, p.level.Height/2, p.level.Length/2);
            //    Vector3Int maxMark = new Vector3Int(p.level.Width/2, p.level.Height/2, p.level.Length/2);
            //    foreach (var face in obj.faces) {
            //        var marks = obj.GetMarks(p.level, face);
            //        foreach (var mark in marks) {
            //            if (mark.X <= minMark.X && mark.Y <= minMark.Y && mark.Z <= minMark.Z) {
            //                minMark = mark;
            //            }
            //            if (mark.X >= maxMark.X && mark.Y >= maxMark.Y && mark.Z >= maxMark.Z) {
            //                maxMark = mark;
            //            }
            //        }
            //    }
            //    //Command.Find("place").Use(p, "red "+minMark.X+" "+minMark.Y+" "+minMark.Z);
            //    //Command.Find("place").Use(p, "blue "+maxMark.X+" "+maxMark.Y+" "+maxMark.Z);
            //    //p.Message("min coords should be  "+minMark.X+" "+minMark.Y+" "+minMark.Z);
            //    //p.Message("max coords should be "+maxMark.X+" "+maxMark.Y+" "+maxMark.Z);
            //    
            //    if (
            //        (maxMark.X >= p.level.Width  || minMark.X < 0) ||
            //        (maxMark.Y >= p.level.Height || minMark.Y < 0) ||
            //        (maxMark.Z >= p.level.Length || minMark.Z < 0)
            //    ) {
            //        p.Message("&cThis model does not fit in the map. Consider lowering the scale argument.");
            //        p.Message("If you would like to import it anyway, use: &T/{0} {1}{2}", name, message, CONFIRMED_STRING);
            //        return;
            //    }           
            //}
            
            ImagePalette palette = ImagePalette.Find(paletteName);
            if (palette == null) { p.Message("&WNo palette matches \"{0}\". &SSee /help img.", paletteName); return; }

            IBitmap2D texture = textureUrl == "" ? null : MCGalaxy.Util.ImageUtils.DecodeImage(HttpUtil.DownloadData(textureUrl, p), p);
            if (texture != null) {
                //p.Message("texture width {0}, texture height {1}", texture.Width, texture.Height);
            } else {
                //p.Message("NULL texture.");
            }

            ImportObjOp op = new ImportObjOp();
            op.texture = texture;
            op.palette = palette;
            op.obj     = obj;
            DrawOpPerformer.Do(op, new SolidBrush(Block.Stone), p, new Vector3Int[1]);
        }
        public override void Help(Player p) {
            p.Message("&T/IO [url] [scale] <url of texture> <palette>");
            p.Message("&HImports a .obj model into the center of the world.");
            p.Message("&T<scale>&H defaults to 16.0");
            p.Message("&T<url of texture>&H optional. Without this argument the object will be made of stone. You must specify a scale to specify this argument.");
            p.Message("&HIf the model is not triangulated, it will have missing faces.");
            p.Message("&HYou can use the free software Blender to import, then export the .obj file triangulated to fix this.");
            //p.Message("&H╨ %7KNOWN ISSUE: you may be disconnected from the server without receiving a kick screen if the model is large.");
            //p.Message("&H╨ %7If your ping disappears and no messages can be sent, you must rejoin the server.");
        }
    }
    

    
    public class Obj 
    {       
        public List<Vector3> vertexCoords = new List<Vector3>();
        public List<Vector2> texCoords = new List<Vector2>();
        
        public List<Face> faces = new List<Face>();
        
        public float scale;
        
        const string vertexPrefix = "v ";
        const string uvPrefix = "vt ";
        const string facePrefix = "f ";
        public Obj (string file, float scale) {
            this.scale = scale;
            string[] lines = file.Split(
                new[] { "\r\n", "\r", "\n" },
                StringSplitOptions.None
            );
            foreach (string line in lines) {
                if (line.StartsWith(vertexPrefix)) { HandleVertex(line); continue; }
                if (line.StartsWith(uvPrefix)) { HandleUv(line); continue; }
                if (line.StartsWith(facePrefix)) { HandleFace(line); continue; }
            }
        }
        //no error handling because we live on the edge
        
        void HandleVertex(string line) {
            line = line.Substring(vertexPrefix.Length);
            string[] coords = line.SplitSpaces();
            float x, y, z;
            Single.TryParse(coords[0], out x); Single.TryParse(coords[1], out y); Single.TryParse(coords[2], out z);
            vertexCoords.Add(new Vector3(x * scale, y * scale, z * scale));
        }
        void HandleUv(string line) {
            line = line.Substring(uvPrefix.Length);
            string[] uvs = line.SplitSpaces();
            float x, y;
            Single.TryParse(uvs[0], out x); Single.TryParse(uvs[1], out y);
            texCoords.Add(new Vector2(x, y));
        }
        void HandleFace(string line) {
            line = line.Substring(facePrefix.Length);
            faces.Add(new Face(line));
        }
        
        public class Face {
            static char[] slashSplit = new char[] { '/' };
            public int[] vertexIndices = new int[3];
            public int[] textureIndices = new int[3];
            
            public Face (string data) {
                // vertexIndex / textureIndex / normalIndex

                // 2/1/1 3/2/1 1/3/1
                string[] infos = data.SplitSpaces();
                
                //we only support triangles so don't go past 3
                for (int i = 0; i < 3; i++) {
                    string[] faceInfo = infos[i].Split(slashSplit);

                    Int32.TryParse(faceInfo[0], out vertexIndices[i]);
                    if (faceInfo.Length > 1) { Int32.TryParse(faceInfo[1], out textureIndices[i]); }
                    
                    //stupid obj indexes start at 1 so if there is none default to 1 instead of 0 which goes out of bounds later since we have to subtract one to access
                    if (textureIndices[i] == 0) { textureIndices[i] = 1; } 
                    
                    CmdIO.printShit = false;
                }
            }
        }
        public Vector3Int[] GetMarks(Level level, Face face) {
            Vector3Int[] marks = new Vector3Int[3];
            for (int i = 0; i < 3; i++) {
                marks[i] = new Vector3Int(
                (int)vertexCoords[face.vertexIndices[i]-1].X + level.Width/2,
                (int)vertexCoords[face.vertexIndices[i]-1].Y + level.Height/2,
                (int)vertexCoords[face.vertexIndices[i]-1].Z + level.Length/2
                );
            }
            return marks;
        }
        public Vector2[] GetUVs(Face face) {
            Vector2[] UVs = new Vector2[3];
            for (int i = 0; i < 3; i++) {
                float x = texCoords.Count > face.textureIndices[i]-1 ? texCoords[face.textureIndices[i]-1].X : 0;
                float y = texCoords.Count > face.textureIndices[i]-1 ? texCoords[face.textureIndices[i]-1].Y : 0;
                UVs[i] = new Vector2(x, y);
            }
            
            // adjust each face's UV inward by a small amount to try to reduce bleed from outside
            Vector2 center = Vector2.Lerp(UVs[0], UVs[1], 0.5f);
            center = Vector2.Lerp(center, UVs[2], 0.5f);
            
            UVs[0] = Vector2.Lerp(UVs[0], center, 0.05f);
            UVs[1] = Vector2.Lerp(UVs[1], center, 0.05f);
            UVs[2] = Vector2.Lerp(UVs[2], center, 0.05f);
            
            return UVs;
        }
    }
    
    public struct Vector2 {
        public float X, Y;
        public Vector2(float X, float Y) {
            this.X = X; this.Y = Y;
        }
        public static Vector2 Lerp(Vector2 a, Vector2 b, float amount) {
            return a + (b - a) * amount;
        }
        
        public static Vector2 operator * (float a, Vector2 b) {
            return new Vector2(a * b.X, a * b.Y);
        }
        public static Vector2 operator * (Vector2 a, float b) {
            return new Vector2(b * a.X, b * a.Y);
        }
        
        public static Vector2 operator - (Vector2 a, Vector2 b) {
            return new Vector2(a.X - b.X, a.Y - b.Y);
        }
        
        public static Vector2 operator + (Vector2 a, Vector2 b) {
            return new Vector2(a.X + b.X, a.Y + b.Y);
        }
        //&l[■] SpiralP: &f6 + (10 - 6)/2
    }
    
    public class ImportObjOp : DrawOp 
    {
        public IBitmap2D texture;
        public ImagePalette palette;
        public Obj obj;
        Vector2[] UVs;

        public override string Name { get { return "ImportObj"; } }
        
        public override long BlocksAffected(Level lvl, Vec3S32[] marks) {
            return -1; // Too hard to estimate blocks affected
        }

        public override void Perform(Vector3Int[] marks, Brush brush, DrawOpOutput output) {
            using (texture)
            {
                if (texture != null) texture.LockBits();

                foreach (var face in obj.faces) 
                {
                    marks = obj.GetMarks(Level, face);
                    UVs   = obj.GetUVs(face);

                    SetMarks(marks);
                    brush.Configure(this, Player);
                    DrawTriangle(marks, brush, output);
                }
            }
        }
        
        void DrawTriangle(Vec3S32[] marks, Brush brush, DrawOpOutput output) {
            Vec3F32 V1 = marks[2], V2 = marks[1], V3 = marks[0];
            Vec3F32 N = Vec3F32.Cross(V2 - V1, V3 - V1);
            N = Vec3F32.Normalise(N);
            Vec3U16 min = Clamp(Min), max = Clamp(Max);
            
            IPaletteMatcher selector = new RgbPaletteMatcher();
            selector.SetPalette(palette.Entries, palette.Entries);
            
            for (ushort yy = min.Y; yy <= max.Y; yy++)
                for (ushort zz = min.Z; zz <= max.Z; zz++)
                    for (ushort xx = min.X; xx <= max.X; xx++)
            {
                // Project point onto the plane
                Vec3F32 P = new Vec3F32(xx, yy, zz);
                float t = Vec3F32.Dot(N, V1) - Vec3F32.Dot(N, P);
                Vec3F32 P0 = P + t * N;
                float dist = (P - P0).Length;
                if (dist > 0.5) continue;
                
                // Check if inside the triangle
                Vec3F32 v0 = V3 - V1, v1 = V2 - V1, v2 = P0 - V1;
                float dot00 = Vec3F32.Dot(v0, v0);
                float dot01 = Vec3F32.Dot(v0, v1);
                float dot02 = Vec3F32.Dot(v0, v2);
                float dot11 = Vec3F32.Dot(v1, v1);
                float dot12 = Vec3F32.Dot(v1, v2);

                // Compute barycentric coordinates
                float invDenom = 1 / (dot00 * dot11 - dot01 * dot01);
                float barry1 = (dot11 * dot02 - dot01 * dot12) * invDenom;
                float barry2 = (dot00 * dot12 - dot01 * dot02) * invDenom;
                float barry3 = 1 - barry1 - barry2;

                
                if (barry1 >= 0 && barry2 >= 0 && barry1 + barry2 <= 1) {
                    if (texture != null) {
                        float texU = barry1*UVs[0].X + barry2*UVs[1].X + barry3*UVs[2].X;
                        float texV = barry1*UVs[0].Y + barry2*UVs[1].Y + barry3*UVs[2].Y;
                        texU = FixedMod(texU, 1.0f);
                        texV = FixedMod(texV, 1.0f);
                        
                        int textureX = (int)(texU * (texture.Width));
                        int textureY = (int)((1 - texV) * (texture.Height));
                        if (textureX >= texture.Width) { textureX = texture.Width-1; } else if (textureX < 0) { textureX = 0; }
                        if (textureY >= texture.Height) { textureY = texture.Height-1; } else if (textureY < 0) { textureY = 0; }
                        
                        
                        Pixel pixel   = texture.Get(textureX, textureY);
                        BlockID block = Block.Air;
                        if (pixel.A >= 127) {
                            block = selector.BestMatch(ref pixel);
                        }
                        brush = new ReplaceBrush(new BlockID[] { 0 }, block);
                    }
                    
                    output(Place(xx, yy, zz, brush));
                }
                
                // /*
                else {
                    Vector2 UVa = new Vector2(0, 0);
                    Vector2 UVb = new Vector2(0, 0);
                    
                    float interp;
                    bool anyOutput = false;
                    if (Axis(P, V1, V2, out interp)) { //use UV1 UV2
                        UVa = UVs[2]; UVb = UVs[1];
                        anyOutput = true;
                    } else 
                    if (Axis(P, V1, V3, out interp)) { //use UV1 UV3
                        UVa = UVs[2]; UVb = UVs[0];
                        anyOutput = true;
                    } else
                    if  (Axis(P, V2, V3, out interp)) { //use UV2 UV3
                        UVa = UVs[1]; UVb = UVs[0];
                        anyOutput = true;
                    }
                    if (anyOutput) {
                        Vector2 resultUV = new Vector2(UVa.X + interp * (UVb.X - UVa.X), UVa.Y + interp * (UVb.Y - UVa.Y) );
                        resultUV.X = FixedMod(resultUV.X, 1.0f);
                        resultUV.Y = FixedMod(resultUV.Y, 1.0f);
                        if (texture != null) {
                            int textureX = (int)(resultUV.X * (texture.Width));
                            int textureY = (int)((1 - resultUV.Y) * (texture.Height));
                            if (textureX >= texture.Width) { textureX = texture.Width-1; } else if (textureX < 0) { textureX = 0; }
                            if (textureY >= texture.Height) { textureY = texture.Height-1; } else if (textureY < 0) { textureY = 0; }
                            
                            Pixel pixel   = texture.Get(textureX, textureY);
                            BlockID block = Block.Air;
                            if (pixel.A >= 127) {
                                block = selector.BestMatch(ref pixel);
                            }
                            brush = new ReplaceBrush(new BlockID[] { 0 }, block);
                        }
                        
                        output(Place(xx, yy, zz, brush));
                    }
                }
                // */
            }
        }
        
        bool Axis(Vec3F32 P, Vec3F32 P1, Vec3F32 P2, out float t) {
            t = 0;
            // Point to line segment test
            float bottom = (P2 - P1).LengthSquared;
            if (bottom == 0) return (P1 - P).Length <= 0.5f;

            t = Vec3F32.Dot(P2 - P1, P - P1) / bottom;
            if (t < 0 || t > 1) return false;
            Vec3F32 proj = P1 + t * (P2 - P1);
            return (P - proj).Length <= 0.5f;
        }
        public static float FixedMod(float x, float m) {
            float r = x % m;
            return r < 0 ? r + m : r;
        }
    }
    
}